from typing import Tuple, Iterator

from typing import (
        Optional,
        Union,
        Generator,
        Any,
        Dict,
        Callable
)

import torch
from torch import Tensor
from torch.utils.data import Dataset
from torch.utils.data import DataLoader, DistributedSampler

def _set_sampler_epoch(dataloader: object, epoch: int) -> None:
    """Calls the ``set_epoch`` method on either the sampler of the given dataloader.

    Every PyTorch dataloader has either a sampler or a batch sampler. If the sampler is wrapped by a
    :class:`~torch.utils.data.distributed.DistributedSampler`, ``set_epoch`` must be called at the beginning
    of every epoch to ensure shuffling applies a new ordering. This has no effect if shuffling is off.

    """
    # cannot use a set because samplers might be unhashable: use a dict based on the id to drop duplicates
    objects: Dict[int, Any] = {}
    # check dataloader.sampler
    if (sampler := getattr(dataloader, "sampler", None)) is not None:
        objects[id(sampler)] = sampler
    # check dataloader.batch_sampler.sampler
    if (batch_sampler := getattr(dataloader, "batch_sampler", None)) is not None and (
        sampler := getattr(batch_sampler, "sampler", None)
    ) is not None:
        objects[id(sampler)] = sampler
    for obj in objects.values():
        set_epoch = getattr(obj, "set_epoch", None)
        if callable(set_epoch):
            set_epoch(epoch)

def apply_to_collection(
    data: Any,
    function: Callable,
    *args: Any,
    **kwargs: Any,
) -> Any:
    
    if data.__class__ is list:
        return [function(x, *args, **kwargs) for x in data]
    if data.__class__ is tuple:
        return tuple(function(x, *args, **kwargs) for x in data)
    if data.__class__ is dict:
        return {k: function(v, *args, **kwargs) for k, v in data.items()}

    assert False
    
def move_data_to_device(batch: Any, device) -> Any:
     
    if isinstance(device, str):
        device = torch.device(device)

    def batch_to(data: Any) -> Any:
        kwargs = {}

        if isinstance(data, Tensor) and isinstance(device, torch.device):
            kwargs["non_blocking"] = True
        data_output = data.to(device, **kwargs)
        if data_output is not None:
            return data_output
        return data
    return apply_to_collection(batch, function=batch_to)

class _DistributonDataLoader:
    def __init__(self, dataloader: DataLoader, device: Optional[torch.device] = None) -> None:
        self.__dict__.update(dataloader.__dict__)
        self._dataloader = dataloader
        self._device = device
        self._num_iter_calls = 0

    def __len__(self) -> int:
        return len(self._dataloader)

    def __iter__(self) -> Union[Iterator[Any], Generator[Any, None, None]]:
        # Without setting the epoch, the distributed sampler would return the same indices every time, even when
        # shuffling is enabled.By doing this the user does not have to call `.set_epoch()` on the sampler.
        _set_sampler_epoch(self._dataloader, self._num_iter_calls)
        self._num_iter_calls += 1

        if self._device is None:
            yield from iter(self._dataloader)
        else:
            for item in self._dataloader:
                yield move_data_to_device(item, self._device)

def collate_fn(
    batch,
    max_length: int = 1024,
) -> Tuple[Tensor, Tensor]:
    
    x = torch.zeros(len(batch), max_length, dtype=torch.long)
    y = torch.zeros(len(batch), max_length, dtype=torch.long)

    for i, encoding in enumerate(batch):
        seq_length = min(len(encoding) - 1, max_length)
        x[i, :seq_length] = encoding[:seq_length]
        y[i, :seq_length] = encoding[1 : seq_length + 1]

    return x, y

def build_dataloader(
                 dataset: Dataset,
                 bs: int,
                 seq_len: int,
                 device: str,
                 distributed: bool,
                 shuffle: bool = True,
    ):
    # TODO: In the future we will need to use our own IterableDataset
    # wit the logic for the distributed computation

    sampler = None

    if distributed:
        sampler = DistributedSampler(dataset, shuffle=shuffle)

    dataloader = DataLoader(
        dataset,
        batch_size=bs,
        sampler=sampler,
        collate_fn=collate_fn,
        drop_last=True,
    )

    if distributed:
        dataloader = _DistributonDataLoader(dataloader=dataloader, device=device)

    return dataloader