import os

from abc import abstractmethod
from typing_extensions import override

class Environment:
    @abstractmethod
    def local_rank(self) -> int:
        """The rank of the currently running process inside of the current node."""

    @abstractmethod
    def main_address(self) -> str:
        """The IP address of the master node."""
    
    @abstractmethod
    def main_port(self) -> str:
        """The port used for communication with the master node."""

    @abstractmethod
    def set_world_size(self, size: int) -> None:
        """Sets the world size."""
    
    @property
    @abstractmethod
    def creates_processes_externally(self) -> bool:
        """Whether the environment creates the subprocesses or not."""

class SingleNodeEnvironment(Environment):

    @override
    def local_rank(self) -> int:
        return int(os.environ.get("LOCAL_RANK", 0))
    
    @override
    def main_address(self) -> str:
        return os.environ.get("MASTER_ADDR", "127.0.0.1")

    @override
    def main_port(self) -> str:
        return os.environ.get("MASTER_PORT", "12355")

    @override
    def set_world_size(self, size: int) -> None:
        self.world_size = size

    @property
    @override
    def creates_processes_externally(self) -> bool:
        """Returns whether the cluster creates the processes or not.

        If at least :code:`LOCAL_RANK` is available as environment variable, Lightning assumes the user acts as the
        process launcher/job scheduler and Lightning will not launch new processes.

        """
        return "LOCAL_RANK" in os.environ

