import logging
import os
import subprocess
import sys

from typing import (
    Any,
    Callable,
    Sequence,
    List
)

from random import getstate as python_get_rng_state
from typing_extensions import override
import torch

from .environment import Environment


def _basic_subprocess_cmd() -> Sequence[str]:
    import __main__ 

    if __main__.__spec__ is None:  # pragma: no-cover
        return [sys.executable, os.path.abspath(sys.argv[0])] + sys.argv[1:]
    return [sys.executable, "-m", __main__.__spec__.name] + sys.argv[1:]

def _num_cpus_available() -> int:
    if hasattr(os, "sched_getaffinity"):
        return len(os.sched_getaffinity(0))

    cpu_count = os.cpu_count()
    return 1 if cpu_count is None else cpu_count

def _suggested_max_num_threads(num_processes: int = 1) -> int:
    if num_processes < 1:
        raise ValueError(f"`num_processes` should be >= 1, got {num_processes}.")
    return max(1, _num_cpus_available() // num_processes)


class Launcher:
    pass

class SubprocessLauncher(Launcher):

    def __init__(
        self,
        environment: Environment,
        num_processes: int,
        num_nodes: int,
    ) -> None:
        self.num_processes = num_processes
        self.num_nodes = num_nodes
        self.environment = environment
        self.procs: List[subprocess.Popen] = []  # launched child subprocesses, does not include the launcher

    def launch(self, function: Callable, *args: Any, **kwargs: Any) -> Any:

        # Launch processes
        print("[INFO] Creating children processes")
        if not self.environment.creates_processes_externally:
            self._call_children_scripts()

        # TODO: launch process observer over self.procs

        # Set number of threads to use
        if "OMP_NUM_THREADS" not in os.environ:
            num_threads = _suggested_max_num_threads(self.num_processes)
            torch.set_num_threads(num_threads)
            os.environ["OMP_NUM_THREADS"] = str(num_threads)

        return function()
        
    def _call_children_scripts(self) -> None:
        # Sanity check that we're the master process and
        # only starting them now
        assert(len(self.procs) == 0)
        assert(self.environment.local_rank() == 0)

        # DDP Environment variables
        os.environ["MASTER_ADDR"] = self.environment.main_address()
        os.environ["MASTER_PORT"] = str(self.environment.main_port())

        print('[INFO] Launching {} processes'.format(self.num_processes))

        for local_rank in range(1, self.num_processes):
            env_copy = os.environ.copy()
            env_copy["LOCAL_RANK"] = f"{local_rank}"

            # remove env var if global seed not set
            if os.environ.get("PL_GLOBAL_SEED") is None and "PL_GLOBAL_SEED" in env_copy:
                del env_copy["PL_GLOBAL_SEED"]

            # start process
            command = _basic_subprocess_cmd()

            stdout_file = open(f'node_{self.environment.local_rank()}.out', 'w')
            stderr_file = open(f'node_{self.environment.local_rank()}.err', 'w')

            proc = subprocess.Popen(command, env=env_copy, stdout=stdout_file, stderr=stderr_file)
            self.procs.append(proc)