import os
from contextlib import nullcontext
from abc import abstractmethod

from typing import (
    Optional,
    List,
    Any
)

from typing_extensions import override


import torch
import torch.nn as nn
from torch import Tensor
from torch.nn.parallel.distributed import DistributedDataParallel

from .environment import Environment
from .launcher import SubprocessLauncher


class Strategy:
    """Strategies for distributed training."""

    def __init__(self, environment):
        self.environment = environment

    @abstractmethod
    def setup_environment(self):
        """Setup the environment for the parallelism strategy."""

    @abstractmethod
    def setup_module(self, module):
        """Setups the model."""

    @abstractmethod
    def setup_optimizer(self, optimizer):
        """Setups the model."""

    @abstractmethod
    def model_to_device(self, model):
        """Moves the model to device."""

    @abstractmethod
    def backwards(self, tensor: Tensor,  *args: Any, **kwargs: Any):
        """Performs a backwards pass."""

    @property
    def local_rank(self) -> int:
        """Returns the local rank"""
        return self.environment.local_rank()


class DataParalelism(Strategy):
    """Implementation for Data Paralellism strategy on a single node"""

    def __init__(self, devices: List[torch.device] = None, environment: Environment = None):
        super().__init__(environment)
        self.environment = environment
        self.devices = devices
        self.num_processes = len(devices)

        # Set the device launcher.
        self.launcher = SubprocessLauncher(self.environment, self.num_processes, 1)

    @property
    def root_device(self) -> torch.device:
        assert self.devices is not None
        return self.devices[self.local_rank]
    
    @override
    def setup_environment(self):

        # Setup the CUDA device
        torch.cuda.set_device(self.root_device)

        # Setup distributed (currently only 1 node)
        self.environment.set_world_size(1 * self.num_processes)
        os.environ["MASTER_ADDR"] = self.environment.main_address()
        os.environ["MASTER_PORT"] = str(self.environment.main_port())

        print("[INFO] Initializing process group with {} and {}".format(self.local_rank, self.environment.world_size))
        torch.distributed.init_process_group("nccl", rank=self.local_rank, world_size=self.environment.world_size)

    @override
    def setup_module(self, module: nn.Module):
        assert module is not None

        device_ids = [self.local_rank]
        ctx = torch.cuda.stream(torch.cuda.Stream()) if device_ids is not None else nullcontext()
        with ctx:
            return DistributedDataParallel(module=module, device_ids=device_ids)

    @override
    def setup_optimizer(self, optimizer):
        return optimizer

    @override
    def model_to_device(self, model: nn.Module) -> None:
        model.to(self.root_device)

    @override
    def backwards(self, tensor: Tensor,  *args: Any, **kwargs: Any):
        tensor.backward(*args, **kwargs)
