import time
import math
from typing import Callable

import torch
import torch.nn as nn
import torch.optim as optim
from fastprogress.fastprogress import master_bar, progress_bar, format_time

from torch.utils.data import DataLoader

from fastprogress.fastprogress import master_bar, progress_bar, format_time


from .distributon import Distributon

class Trainer():

    def __init__(self,
                 train_loader: DataLoader,
                 valid_loader: DataLoader,
                 model: nn.Module,
                 criterion: Callable,
                 optimizer: optim.Optimizer,
                 lr: float,
                 grad_acc_steps: int,
                 grad_clip_norm: float,
                 model_save_path: str,
                 use_fp16: bool,
                 distributon:Distributon = None,
                 progress_bar: bool = True):
        """
        Trainer implementation for Optimus models.

        Args:
            dl (OptimusDataLoader): Dataloader used to train the model.
            model (nn.Module): Model to train.
            criterion (callable): A suitable loss function. The trainer assumes
                nn.CrossEntropyLoss, though other loss funcs (like
                label-smoothed cross entropy loss) can be used.
            optimizer (torch.optim.Optimizer): Optimizer to use for training.
            lr (float): Max learning rate value to use for one-cycle scheduling.
            grad_acc_steps (int): Number of gradient accumulation steps before
                running backpropagation. Setting to 1 effectively disables
                gradient accumulation.
            grad_clip_norm (float): Gradient clipping norm value.
            model_save_path (str): The best model (based on validation loss) is
                saved to the specified path.
            use_fp16 (bool): Whether to train the model in 16-bit floating point
                precision. If such hardware is not supported, a warning is
                issued and normal 32-bit precision is used instead.
            progress_bar (bool): Whether to show a progress bar in console while
                training. This is automatically disabled if output is a file,
                however some stats are printed after finishing epochs. If False,
                no stats are printed at all during training, whether the output
                is a console or a file.

        """
        self.train_loader = train_loader
        self.valid_loader = valid_loader
        self.model = model
        self.criterion = criterion
        self.optimizer = optimizer
        self.lr = lr


        self.distributon = distributon

        if distributon is not None:
            self.model = self.distributon.setup_model(model)
            self.optimizer = self.distributon.setup_optimizer(optimizer)

        assert type(grad_acc_steps) is int and grad_acc_steps > 0
        self.grad_acc_steps = grad_acc_steps

        self.grad_clip_norm = grad_clip_norm
        self.model_save_path = model_save_path

        self.use_fp16 = use_fp16
        self.fp16_dtype = torch.float16

        self.progress_bar = progress_bar

    def fit(self, n_epochs: int) -> None:
        """
        Fit the model, using the trainer, on the data inside the dataloader
        object.

        Args:
            n_epochs (int): Number of epochs to train for.

        """
        # this is fastai's implementation of one cycle
        self.scheduler = torch.optim.lr_scheduler.OneCycleLR(
            optimizer=self.optimizer,
            max_lr=self.lr,
            epochs=n_epochs,
            steps_per_epoch=len(self.train_loader) // self.grad_acc_steps)

        # scaler used for mixed precision fp16 training on GPU
        self.scaler = torch.cuda.amp.GradScaler(enabled=self.use_fp16)

        best_val_loss = float('inf')

        # progress bar for epochs
        self.mb = master_bar(list(range(n_epochs)))

        # start training for n_epochs
        for self.epoch in range(n_epochs):

            if self.progress_bar is True:
                self.mb.update(self.epoch)

            start_time = time.time()
            self._do_epoch_train()
            self._do_epoch_validate()
            self.epoch_time = time.time() - start_time

            # write end of epoch stats
            self._write_epoch_stats()

            # if better model on validation loss, save it
            if self.val_loss < best_val_loss:
                best_val_loss = self.val_loss
                torch.save(self.model.state_dict(), self.model_save_path)

    def _do_epoch_train(self):
        self.model.train() # put model in training mode

        # compute average train loss, train perplexity and ms/batch every ~200
        # batches, or every 10% of training dataset (whichever is smaller),
        # rounded to gradient accumulation steps
        est_interval = int(max(min(200, 0.1 * len(self.train_loader)) // self.grad_acc_steps, 1) * self.grad_acc_steps)

        # progress bar for batches
        pb = progress_bar(range(len(self.train_loader)), parent=self.mb)

        self.ms_per_batch = 0.
        total_loss = 0.
        start_time = time.time()

        for i, (x, y) in enumerate(self.train_loader):

            if self.progress_bar is True:
                pb.update(i)

            # automatic mixed precision training
            with torch.cuda.amp.autocast(dtype=self.fp16_dtype,
                                         enabled=self.use_fp16):
                output = self.model(x)
                loss = self.criterion(output.view(-1, len(self.train_loader.dataset.tok)),
                                  y.reshape(-1))

                loss = loss / self.grad_acc_steps # normalize to account for gradient accumulation

            if self.distributon:
                self.distributon.backward(loss)
            else:
                self.scaler.scale(loss).backward()

            total_loss += loss.item()

            # update only after grad_acc_steps
            if (i + 1) % self.grad_acc_steps == 0:
                # Note: This will ignore the last few batches of the dataset,
                # when the gradient accumulation steps are more than 1, and the
                # number of batches doesn't cleanly divide by grad_acc_steps

                if not self.distributon:
                    # gradient clipping
                    self.scaler.unscale_(self.optimizer)

                nn.utils.clip_grad_norm_(self.model.parameters(),
                                         max_norm=self.grad_clip_norm)

                # TODO: Fix this for distributed
                if not self.distributon:
                    self.scaler.step(self.optimizer)
                    self.scaler.update()

                self.optimizer.zero_grad()

                self.scheduler.step()
                lr = self.scheduler.get_last_lr()[0]

                # update train loss, train ppl and estimated ms/batch
                if (i + 1) % est_interval == 0:
                    self.ms_per_batch = (time.time() - start_time) * 1000 / est_interval
                    self.train_loss = (total_loss * self.grad_acc_steps) / est_interval
                    self.train_ppl = math.exp(self.train_loss)

                    total_loss = 0.
                    start_time = time.time()

                self.mb.child.comment = f" | train loss: {loss.item() * self.grad_acc_steps:.4f} | " \
                                        f"~{self.ms_per_batch:.2f} ms/batch | " \
                                        f" lr: {lr:.7f}"

        pb.on_iter_end()

    def _do_epoch_validate(self):
        self.model.eval() # put model in eval mode

        total_loss = 0.

        # progress bar for batches
        pb = progress_bar(range(len(self.valid_loader)), parent=self.mb)

        with torch.no_grad():
            for i, (x, y) in enumerate(self.valid_loader):

                if self.progress_bar is True:
                    pb.update(i)

                with torch.cuda.amp.autocast(dtype=self.fp16_dtype,
                                             enabled=self.use_fp16):
                    output = self.model(x)
                    loss = self.criterion(output.view(-1, len(self.valid_loader.dataset.tok)),
                                          y.reshape(-1))

                total_loss += loss.item()

                self.mb.child.comment = f" | valid loss: {loss.item():.4f}"

            self.val_loss = total_loss / (len(self.valid_loader) - 1)
            self.val_ppl = math.exp(self.val_loss)

            pb.on_iter_end()

    def _write_epoch_stats(self):
        if self.progress_bar is True:
            epoch_time = format_time(self.epoch_time)
            self.mb.write(
                f"* End of epoch {self.epoch:3d}:\n"
                f"\tTotal time: {epoch_time:9s} | "
                f"Est. ms/batch: {self.ms_per_batch:.2f}\n"
                f"\tTotal train batches: {len(self.train_loader):10d} | "
                f"Train loss: {self.train_loss: 7.2f} | "
                f"Train perplexity: {self.train_ppl: 8.2f}\n"
                f"\tTotal valid batches: {len(self.valid_loader):10d} | "
                f"Valid loss: {self.val_loss: 7.2f} | "
                f"Valid perplexity: {self.val_ppl: 8.2f}")
