import fire
import torch
from torch import nn

from optimus.datasets import TinyStoriesDataset
from optimus.tokenizers import SentencePieceTokenizer
from optimus.distributon.dataloader import build_dataloader
from optimus.models import OptimusTransformer
from optimus.trainer import Trainer

from optimus.distributon import Distributon

def main(batch_size: int = 8,
         grad_acc_steps: int = 1,
         seq_len: int = 512,
         lr_max: float = 1e-4,
         grad_clip_norm: float = 1.0,
         epochs: int = 1,
         tokenizer_path: str = 'optimus.model',
         checkpoints_path: str = 'best_model.pth',
         dim: int = 512,
         n_layers: int = 6,
         n_heads: int = 8,
         dropout: float = 0.0,
         use_fp16: bool = True,
         distributed: bool = False):
    """
    Run the main training loop for the model.

    Args:
        batch_size (int): Batch size for training.
        grad_acc_steps (int): Number of batches to accumulate gradients for
            before running backpropagation to update weights.
        seq_len (int): Context length for training.
        lr_max (float): Maximum learning rate, used for one-cycle scheduling.
        grad_clip_norm (float): Gradient clipping value for gradient's norm.
        epochs (int): Number of epochs to train for.
        tokenizer_path (str): Path to the tokenizer model.
        checkpoints_path (str): Where to save the trained model. Should be a .pt
            or .pth file.
        dim (int): Dimension of the model.
        n_layers (int): Number of layers for the model.
        n_heads (int): Number of heads inside an attention layer for the model.
        dropout (float): Dropout to use for the model.
        use_fp16 (bool): Whether to train using floating-point 16-bits
            precision.

    """

    print(f"Running with:\n"
        f"\t- batch size: {batch_size}\n"
        f"\t- gradient accumulation steps: {grad_acc_steps}\n"
        f"\t- context length: {seq_len}\n"
        f"\t- max learning rate: {lr_max}\n"
        f"\t- gradient clipping norm: {grad_clip_norm}\n"
        f"\t- epochs: {epochs}\n"
        f"\t- tokenizer: {tokenizer_path}\n"
        f"\t- checkpoints path: {checkpoints_path}\n"
        f"\t- model dimension: {dim}\n"
        f"\t- model layers: {n_layers}\n"
        f"\t- model attention heads: {n_heads}\n"
        f"\t- model dropout: {dropout}\n"
        f"\t- 16-bit floating-point training (fp16): {use_fp16}\n"
        f"Please see '--help' if you want to change these settings")

    
    # Launch the distributed proccesses
    if distributed:
        distributon = Distributon([f"cuda:{i}" for i in range(torch.cuda.device_count())])
        distributon.launch()
        device = distributon._strategy.root_device
    else:
        distributon = None

    # load tokenizer
    tok = SentencePieceTokenizer(model_path=tokenizer_path)

    # load dataset splits
    train_ds = TinyStoriesDataset(split='train', tokenizer=tok)
    test_ds = TinyStoriesDataset(split='test', tokenizer=tok)

    print(f"Number of examples in training set: {len(train_ds)}")
    print(f"Number of examples in testing set: {len(test_ds)}")

    # create the dataloaders
    train_loader = build_dataloader(train_ds,
                           bs=batch_size,
                           seq_len=seq_len,
                           device=device,
                           distributed=distributed)

    valid_loader = build_dataloader(train_ds,
                        bs=batch_size,
                        seq_len=seq_len,
                        device=device,
                        distributed=distributed)

    # create model and move to device
    model = OptimusTransformer(len(tok),
                               n_layers=n_layers,
                               dim=dim,
                               n_heads=n_heads,
                               p_drop=dropout,
                               weight_tying=False)

    if not distributed:
        model = model.to('cuda')

    _total_params = sum(p.numel() for p in model.parameters())
    print(f"Number of model parameters: {_total_params}")

    # define loss metric
    criterion = nn.CrossEntropyLoss()

    # define optimizer
    # see [1] for a discussion on what the epsilon value should be for amp; 1e-7
    # is a good default for both amp and normal training
    # [1]: https://github.com/pytorch/pytorch/issues/26218
    optimizer = torch.optim.Adam(model.parameters(), betas=(0.9, 0.999), eps=1e-7)

    print("Starting training...")

    # create trainer and start fitting
    trainer = Trainer(train_loader=train_loader,
                      valid_loader=valid_loader,
                      model=model,
                      criterion=criterion,
                      optimizer=optimizer,
                      lr=lr_max,
                      grad_acc_steps=grad_acc_steps,
                      grad_clip_norm=grad_clip_norm,
                      model_save_path=checkpoints_path,
                      use_fp16=use_fp16,
                      distributon=distributon,
                      progress_bar=True)
    trainer.fit(epochs)

    print(f"Finished training! Best model weights saved at '{checkpoints_path}'")


if __name__=="__main__":
    fire.Fire(main)
